#include <cv.h>
//#include <cxcore.h>
#include <highgui.h>
#include <stdio.h>
#include <stdlib.h>

int n_boards = 0;		//Number of snapshots of the chessboard
int frame_step;			//Frames to be skipped
int board_w;			//Enclosed corners horizontally on the chessboard
int board_h;			//Enclosed corners vertically on the chessboard
int main(int argc, char* argv[]) 
{
   CvCapture* capture;
   char *fintr;
   char *fdist;
   //sprintf(fintr,"Intrinsics.xml");
   fintr = "Intrinsics.xml";
   fdist = "Distortion.xml";
   //   char foutlogname[1024];
   //   sprintf(foutlogname,"%s%s",*(argv + foutno),".log");

   board_w  = 8;
   board_h  = 6;
   int f_in = 0;
   int f_intr = 0;
   int f_dist = 0;

   for (int i=1; i<argc; i++){
      if(0 == strcmp(*(argv+i), "-i"))          f_in = i+1;
      if(0 == strcmp(*(argv+i), "-oi"))         f_intr = i+1;
      if(0 == strcmp(*(argv+i), "-od"))         f_dist = i+1;
      if(0 == strcmp(*(argv+i), "-w"))          board_w = atoi(*(argv+i+1));
      if(0 == strcmp(*(argv+i), "-h"))          board_h = atoi(*(argv+i+1));
   }

   fprintf(stderr,"\n\n**  Written by Dimitris Skarlatos @ Cyprus university of Technology. v.0.1.1\n");
   if ((f_in==0) ){
      fprintf(stderr,"************************************************************\n");
      fprintf(stderr,"Usage:\n-i input_video.avi\n");
      fprintf(stderr,"-oi Intrinsics_filename. Default Intrinsics.xml\n");
      fprintf(stderr,"-od distortion_filename. Default Distortion.xml\n");
      fprintf(stderr,"-w board width. Refers to the boxes minus one Default = 6\n");
      fprintf(stderr,"-h board height. Refers to the boxes minus one Default = 8\n");

      fprintf(stderr,"Useful notes\n");
      fprintf(stderr,"The boxes might be any number (preferably not equal)\n");
      fprintf(stderr,"The program will ask for frames the user wants to use for calibration. It will then divide the total number of frames by this number and use one frame every x frames\n");
      fprintf(stderr,"The video input can read most avis and some other format depending on OpenCV formats\n");
      fprintf(stderr,"Check www.photogrammetric-vision.weebly.com on how to convert videos in readable format\n");
      fprintf(stderr,"Output is two files, one with intrinsic parameters and the other one with distortion coefficients.\n");
      exit(0);
   }


   if (board_w < board_h) { int lte = board_h; board_h = board_w; board_w = lte; } 
   int board_total  = board_w * board_h;						//Total enclosed corners on the board
   CvSize board_sz = cvSize( board_w, board_h );

   //capture = cvCreateCameraCapture( 0 ); // for onboard video camera
   //capture = cvCreateFileCapture( "calibration_seq.avi");
   capture = cvCreateFileCapture( *(argv+f_in));
   if(!capture) 
   { 
      fprintf(stderr,"\nCouldn't open the video file\n"); 
      fprintf(stderr,"OpenCV cannot read this video\n");
      fprintf(stderr,"Read how to convert it in a compatible format for www.photogrammetric-vision.weebly.com"); 
      return -1;
   }

   /* Read the video's frame size out of the AVI. */
   //CvSize frame_size;
   //frame_size.height =
   //    (int) cvGetCaptureProperty( capture, CV_CAP_PROP_FRAME_HEIGHT );
   //frame_size.width =
   //    (int) cvGetCaptureProperty( capture, CV_CAP_PROP_FRAME_WIDTH );
   //Get number of frames
   cvSetCaptureProperty( capture, CV_CAP_PROP_POS_AVI_RATIO, 1. );
   int number_of_frames = (int) cvGetCaptureProperty( capture, CV_CAP_PROP_POS_FRAMES );
   cvSetCaptureProperty( capture, CV_CAP_PROP_POS_FRAMES, 0. );
   fprintf(stderr, "Number of frames on video= %d\n", number_of_frames);


   printf("Enter the numbers of frames you want to use for caliration = ");
   scanf("%d",&n_boards);

   //printf("\rEnter the numbers of frames to skip = ");
   //scanf("%d",&frame_step);
   frame_step = (int) ( number_of_frames / n_boards);

   cvNamedWindow( "Snapshot" );
   cvNamedWindow( "Raw Video");

   //Allocate storage for the parameters according to total number of corners and number of snapshots
   CvMat* image_points      = cvCreateMat(n_boards*board_total,2,CV_32FC1);
   CvMat* object_points     = cvCreateMat(n_boards*board_total,3,CV_32FC1);
   CvMat* point_counts      = cvCreateMat(n_boards,1,CV_32SC1);
   CvMat* intrinsic_matrix  = cvCreateMat(3,3,CV_32FC1);
   CvMat* distortion_coeffs = cvCreateMat(4,1,CV_32FC1);

   //Note:
   //Intrinsic Matrix - 3x3			   Lens Distorstion Matrix - 4x1
   //	[fx 0 cx]							[k1 k2 p1 p2   k3(optional)]
   //	[0 fy cy]
   //	[0  0  1]


   CvPoint2D32f* corners = new CvPoint2D32f[ board_total ];
   int corner_count;
   int successes = 0;
   int step, frame = 0;

   IplImage *image = cvQueryFrame( capture );
   IplImage *gray_image = cvCreateImage(cvGetSize(image),8,1);			//subpixel


   //Loop while successful captures equals total snapshots
   //Successful captures implies when all the enclosed corners are detected from a snapshot

   while(successes < n_boards)					
   {
      if((frame++ % frame_step) == 0)									//Skip frames
      {
         //Find chessboard corners:
         int found = cvFindChessboardCorners(image, board_sz, corners, &corner_count,CV_CALIB_CB_ADAPTIVE_THRESH | CV_CALIB_CB_FILTER_QUADS );

         cvCvtColor(image, gray_image, CV_BGR2GRAY);										//Get Subpixel accuracy on those corners
         cvFindCornerSubPix(gray_image, corners, corner_count, cvSize(11,11),cvSize(-1,-1), cvTermCriteria(CV_TERMCRIT_EPS+CV_TERMCRIT_ITER, 30, 0.1 ));

         cvDrawChessboardCorners(image, board_sz, corners, corner_count, found);			//Draw it

         // If we got a good board, add it to our data
         if( corner_count == board_total ) 
         {
            cvShowImage( "Snapshot", image );										//show in color if we did collect the image
            step = successes*board_total;
            for( int i=step, j=0; j<board_total; ++i,++j ) {
               CV_MAT_ELEM(*image_points, float,i,0) = corners[j].x;
               CV_MAT_ELEM(*image_points, float,i,1) = corners[j].y;
               CV_MAT_ELEM(*object_points,float,i,0) = (float) j/board_w;
               CV_MAT_ELEM(*object_points,float,i,1) = (float) (j%board_w);
               CV_MAT_ELEM(*object_points,float,i,2) = 0.0f;
            }
            CV_MAT_ELEM(*point_counts, int,successes,0) = board_total;    
            successes++;
            printf("\r%d successful Snapshots out of %d collected.",successes,n_boards);
         }
         else
            cvShowImage( "Snapshot", gray_image );										//Show Gray if we didn't collect the image
      } 


      //Handle pause/unpause and ESC
      int c = cvWaitKey(15);
      if(c == 'p')
      {  
         c = 0;
         while(c != 'p' && c != 27)
         {
            c = cvWaitKey(250);
         }
      }
      if(c == 27)
         return 0;

      image = cvQueryFrame( capture );								//Get next image
      cvShowImage("Raw Video", image);
   } 

   //End WHILE loop with enough successful captures

   cvDestroyWindow("Snapshot");

   printf("\n\n *** Calbrating the camera now...\n");

   //Allocate matrices according to successful number of captures
   CvMat* object_points2  = cvCreateMat(successes*board_total,3,CV_32FC1);
   CvMat* image_points2   = cvCreateMat(successes*board_total,2,CV_32FC1);
   CvMat* point_counts2   = cvCreateMat(successes,1,CV_32SC1);

   //Tranfer the points to matrices
   for(int i = 0; i<successes*board_total; ++i)
   {
      CV_MAT_ELEM( *image_points2, float, i, 0)  =	CV_MAT_ELEM( *image_points, float, i, 0);
      CV_MAT_ELEM( *image_points2, float,i,1)    =	CV_MAT_ELEM( *image_points, float, i, 1);
      CV_MAT_ELEM(*object_points2, float, i, 0)  =  CV_MAT_ELEM( *object_points, float, i, 0) ;
      CV_MAT_ELEM( *object_points2, float, i, 1) =  CV_MAT_ELEM( *object_points, float, i, 1) ;
      CV_MAT_ELEM( *object_points2, float, i, 2) =  CV_MAT_ELEM( *object_points, float, i, 2) ;
   } 

   for(int i=0; i<successes; ++i)
   { 
      CV_MAT_ELEM( *point_counts2, int, i, 0) =  CV_MAT_ELEM( *point_counts, int, i, 0);			
      //These are all the same number
   }
   cvReleaseMat(&object_points);
   cvReleaseMat(&image_points);
   cvReleaseMat(&point_counts);

   // Initialize the intrinsic matrix with both the two focal lengths in a ratio of 1.0

   CV_MAT_ELEM( *intrinsic_matrix, float, 0, 0 ) = 1.0f;
   CV_MAT_ELEM( *intrinsic_matrix, float, 1, 1 ) = 1.0f;
   //CV_MAT_ELEM( *intrinsic_matrix, float, 0, 2 ) = (float) (frame_size.width / 2);
   //CV_MAT_ELEM( *intrinsic_matrix, float, 1, 2 ) = (float) (frame_size.height / 2);

   //Calibrate the camera
   //_____________________________________________________________________________________

   cvCalibrateCamera2(object_points2, image_points2, point_counts2,  cvGetSize( image ),
      intrinsic_matrix, distortion_coeffs, NULL, NULL, CV_CALIB_FIX_ASPECT_RATIO + CV_CALIB_FIX_PRINCIPAL_POINT);
   //CV_CALIB_FIX_ASPECT_RATIO

   //_____________________________________________________________________________________

   //Save values to file
   printf(" *** Calibration Done!\n\n");
   printf("Storing Intrinsics and Distortions files...\n");
   if (f_intr != 0) { fintr = *(argv+f_intr); } 
   if (f_dist != 0) { fdist = *(argv+f_dist); } 
   //cvSave("Intrinsics.xml",intrinsic_matrix);
   //cvSave("Distortion.xml",distortion_coeffs);
   cvSave(fintr , intrinsic_matrix);
   cvSave(fdist , distortion_coeffs);
   printf("Files saved.\n\n");

   printf("Starting corrected display....");

   //Sample: load the matrices from the file
   //CvMat *intrinsic = (CvMat*)cvLoad("Intrinsics.xml");
   //CvMat *distortion = (CvMat*)cvLoad("Distortion.xml");
   CvMat *intrinsic = (CvMat*)cvLoad(fintr);
   CvMat *distortion = (CvMat*)cvLoad(fdist);

   // Build the undistort map used for all subsequent frames.

   IplImage* mapx = cvCreateImage( cvGetSize(image), IPL_DEPTH_32F, 1 );
   IplImage* mapy = cvCreateImage( cvGetSize(image), IPL_DEPTH_32F, 1 );
   cvInitUndistortMap(intrinsic,distortion,mapx,mapy);

   // Run the camera to the screen, showing the raw and the undistorted image.

   cvNamedWindow( "Undistort" );
   while(image) 
   {
      IplImage *t = cvCloneImage(image);
      cvShowImage( "Raw Video", image );			// Show raw image
      cvRemap( t, image, mapx, mapy );			// Undistort image
      cvReleaseImage(&t);
      cvShowImage("Undistort", image);			// Show corrected image

      //Handle pause/unpause and ESC
      int c = cvWaitKey(15);
      if(c == 'p')
      { 
         c = 0;
         while(c != 'p' && c != 27)
         {
            c = cvWaitKey(250);
         }
      }
      if(c == 27)
         break;

      image = cvQueryFrame( capture );
   } 

   return 0;
}



